"""

True parameters
===============

Value of the parameters used to generate synthetic choices.
"""

# %%
SCALE = 0.5
true_parameters = {
    'beta_rating': 1.5 * SCALE,
    'beta_price': -0.8 * SCALE,
    'beta_chinese': 1.5 * SCALE,
    'beta_japanese': 2.5 * SCALE,
    'beta_korean': 1.5 * SCALE,
    'beta_indian': 2 * SCALE,
    'beta_french': 1.5 * SCALE,
    'beta_mexican': 2.5 * SCALE,
    'beta_lebanese': 1.5 * SCALE,
    'beta_ethiopian': 1 * SCALE,
    'beta_log_dist': -1.2 * SCALE,
    'mu_asian': 2.0,
    'mu_downtown': 2.0,
}
